module Data.Louvain where

import Data.List (maximumBy,nub)
import Data.Graph.Inductive

type Modularity = Double

-- | group nodes and produce a new graph 
inducedGraph :: (Eq b, Ord b, DynGraph gr) => gr a b -> (Node, [Node]) -> gr a b
inducedGraph gr (a,b) = delNodes b (insEdges newEdges gr')
    where
        gr' = undir gr
        newEdges = Prelude.map (\(n,l) -> (a,n,l)) ( nub $ concat $ Prelude.map (lsuc gr') b )

-- | TODO exducedGraph (inverse for tests with quickCheck)
inducedGraph' :: (Ord b, DynGraph gr) => gr a b -> [(Node, [Node])] -> gr a b
inducedGraph' gr ns = Prelude.foldl (\gr' n -> inducedGraph gr' n) gr ns

neighbors'' :: DynGraph gr => gr a b -> [Node] -> [Node]
neighbors'' gr ns = nub $ concat (Prelude.map (neighbors gr) ns)

neighcom :: DynGraph gr => gr a b -> Node -> [[Node]]
neighcom gr n = scanl (\x y -> take 1 x ++ [y]) [n] (neighbors gr n)

neighcom' :: DynGraph gr => gr a b -> [Node] -> [[Node]]
neighcom' gr ns = scanl (\x y -> take 1 x ++ [y]) ns (neighbors'' gr ns)

rotate :: Int -> [a] -> [a]
rotate _ [] = []
rotate n xs = zipWith const (drop n (cycle xs)) xs

takeDrop :: Int -> [a] -> [[a]]
takeDrop n xs = [ (take n xs), drop n xs]

-- http://stackoverflow.com/questions/35388734/list-partitioning-implemented-recursively
separate :: [a] -> [[[a]]]
separate [] = [[]]
separate (x:xs) = let recur = separate xs
                      split = do
                        partition <- recur
                        return $ [x] : partition
                      noSplit = do
                        (y:ys) <- recur
                        return $ (x:y):ys
                  in split ++ noSplit


separate' :: forall a. [a] -> [[[a]]]
separate' xs = [ takeDrop t (rotate r xs) 
               | t <- [1.. fromIntegral (length xs) - 1 ]
               , r <- [0.. fromIntegral (length xs)     ]
               ]


gpartition :: DynGraph gr => gr a b -> [[[Node]]]
gpartition gr = separate (nodes gr)

modularities :: DynGraph gr => gr a b -> [[Node]] -> Double
modularities gr xs = sum $ Prelude.map (\y -> modularity gr y) xs

compareModularities :: DynGraph gr => gr a b -> [[Node]] -> [[Node]] -> Ordering
compareModularities gr xs ys
    | modularities gr xs < modularities gr ys = LT
    | modularities gr xs > modularities gr ys = GT
    | otherwise = EQ

bestPartition :: DynGraph gr => gr a b -> [[Node]]
bestPartition gr = maximumBy (compareModularities gr) $ gpartition gr

modularity :: DynGraph gr => gr a b -> [Node] -> Double
modularity gr ns = coverage - edgeDensity
    where
        coverage :: Double
        coverage  = sizeSubGraph / sizeAllGraph 
            where
                sizeSubGraph :: Double
                sizeSubGraph = fromIntegral ( size $ subgraph ns gr )

                sizeAllGraph :: Double
                sizeAllGraph = fromIntegral (size gr)

        edgeDensity :: Double
        edgeDensity = (sum (Prelude.map (\node -> (degree node) / links ) ns)) ** 2
            where
                degree :: Node -> Double
                degree node = fromIntegral (deg gr node)

                links :: Double
                links       = fromIntegral (2 * (size gr))

